#include "module.h"
#include "compiler.h"
#include "variable.h"
#include "function.h"
#include "typedef.h"
#include "enum.h"

#include <frost_utils_file.h>

using namespace Frost;

Module::Module(s_ptr<Compiler> pCompiler, const s_str& sFile) : sFile_(sFile), pCompiler_(pCompiler)
{
}

void Module::SetSelf(s_wptr<Module> pSelf)
{
    pSelf_ = pSelf;
}

void Module::CreateMainClass()
{
    pMainClass_ = s_refptr<Class>(new Class(pCompiler_, pSelf_, "class"));
    lClassStack_.PushBack(pMainClass_);
    pCurrentClass_ = pMainClass_;
}

void Module::SetDocElements(const s_ctnr<s_str>& lDocElements)
{
    if (pCurrentClass_)
        pCurrentClass_->SetDocElements(lDocElements);
}

void Module::SetName(const TypeName& mName)
{
    if (pCurrentClass_ == pMainClass_)
    {
        if (pMainClass_)
        {
            if (pMainClass_->GetType().sType.IsEmpty())
            {
                if (!sFile_.IsEmpty())
                {
                    s_str sFileName = MakeFileName(mName);
                    s_str sRealFile = sFile_.Cut("/").Back();
                    if (sRealFile != sFileName)
                        pCompiler_->ThrowWarning("module name doesn't match file name (expected : "+sFileName+", got : "+sRealFile+")");
                }

                mName_ = mName;
                mName_.pClass = pMainClass_;

                pMainClass_->SetType(mName);
            }
            else
                pCompiler_->ThrowWarning("module name can only be set once, ignoring");
        }
        else
        {
            if (mName_.sType.IsEmpty())
            {
                if (!sFile_.IsEmpty())
                {
                    s_str sFileName = MakeFileName(mName);
                    s_str sRealFile = sFile_.Cut("/").Back();
                    if (sRealFile != sFileName)
                        pCompiler_->ThrowWarning("module name doesn't match file name (expected : "+sFileName+", got : "+sRealFile+")");
                }

                mName_ = mName;
            }
            else
                pCompiler_->ThrowWarning("module name can only be set once, ignoring");
        }
    }
    else
        pCompiler_->ThrowError("cannot set the name of a nested class with the 'module' command");


    pCurrentContext_ = &mGlobalContext_;

    s_ctnr<s_str>::iterator iterNsp;
    foreach (iterNsp, mName_.lNamespaces)
    {
        pCurrentContext_->lContextList[*iterNsp] = Context(pCurrentContext_);
        pCurrentContext_ = pCurrentContext_->lContextList[*iterNsp];
    }

    if (pMainClass_)
    {
        pCurrentContext_->lContextList[mName.sType] = Context(pCurrentContext_);
        pCurrentContext_ = pCurrentContext_->lContextList[mName.sType];
    }
}

void Module::SetHeaderFile(const s_str& sHeader)
{
    sHeaderFile_ = sHeader;
}

void Module::SetPreloaded()
{
    bPreloaded_ = true;
}

const s_bool& Module::IsPreloaded()
{
    return bPreloaded_;
}

void Module::CompileScope_(Scope& s, const s_ctnr<s_str>& lContext)
{
    s_ctnr<Variable>::iterator iterVar;
    foreach (iterVar, s.lVariableList_)
    {
        DecorateType(iterVar->mType, lContext);
        iterVar->mType.Qualify();
        CheckType_(iterVar->mType);
    }

    s_ctnr<Function>::iterator iterFunc;
    foreach (iterFunc, s.lFunctionList_)
    {
        DecorateType(iterFunc->mReturnType, lContext);
        iterFunc->mReturnType.Qualify();
        CheckType_(iterFunc->mReturnType);

        s_ctnr<Variable>::iterator iterVar;
        foreach (iterVar, iterFunc->lParams)
        {
            DecorateType(iterVar->mType, lContext);
            iterVar->mType.Qualify();
            CheckType_(iterVar->mType);
        }

        if (!iterFunc->lDependencies.IsEmpty())
        {
            s_ctnr<TypeName>::const_iterator iterDeps;
            foreach (iterDeps, iterFunc->lDependencies)
            {
                s_bool bFound;
                s_ctnr<s_wptr<Module>>::iterator iterModule;
                foreach (iterModule, lUsedModuleList_)
                {
                    if ((*iterModule)->GetFullName() == iterDeps->sFullType)
                    {
                        bFound = true;
                        HeaderNeedsModule_(*iterModule);
                        break;
                    }
                }

                if (!bFound)
                    pCompiler_->ThrowError("module '"+iterDeps->sFullType+"' has not been previously declared as used");
            }
        }
    }

    s_ctnr<Typedef>::iterator iterType;
    foreach (iterType, s.lTypedefList_)
    {
        if (iterType->bFunction)
        {
            DecorateType(iterType->mReturnType, lContext);
            iterType->mReturnType.Qualify();
            CheckType_(iterType->mReturnType);

            if (!iterType->mMembership.sType.IsEmpty())
            {
                DecorateType(iterType->mMembership, lContext);
                CheckType_(iterType->mMembership);
            }

            s_ctnr<TypeName>::iterator iterParams;
            foreach (iterParams, iterType->lParams)
            {
                DecorateType(*iterParams, lContext);
                iterParams->Qualify();
                CheckType_(*iterParams);
            }
        }
        else
        {
            DecorateType(iterType->mOldType, lContext);
            iterType->mOldType.Qualify();
            CheckType_(iterType->mOldType);
        }
    }
}

void Module::Compile()
{
    s_ctnr<s_str> lContext = mName_.lNamespaces;
    CompileScope_(mGlobalScope_, lContext);

    if (pMainClass_)
        pMainClass_->CompileScopes(lContext);
    else
    {
        s_ctnr<s_refptr<Class>>::iterator iterClass;
        foreach (iterClass, mGlobalScope_.lClassList_)
            (*iterClass)->CompileScopes(lContext);
    }

    if (pCompiler_->IsVerbose()) Log("Serializing ...");
    pCompiler_->CheckOutputDirectory();
    Serialize();
    if (pCompiler_->IsVerbose()) Log("Done.");

    bPreloaded_ = false;
}

const s_str& Module::GetName() const
{
    return mName_.sType;
}

const s_str& Module::GetFullName() const
{
    return mName_.sFullType;
}

void Module::HeaderNeedsModule_(s_wptr<Module> pModule)
{
    if (!lExternalModuleList_.Find(pModule->GetFullName()))
    {
        lExternalModuleList_.PushBack(pModule->GetFullName());
        lExternalModuleList_.Sort();
    }
}

s_str Module::MakeFileName(const TypeName& mName)
{
    s_str sName = mName.sFullType;
    sName.Replace("::", "_");
    sName.ToLower();
    sName += ".mpp";

    return sName;
}

s_str Module::GetCreatedSourceFile() const
{
    s_str sSource = GetFullName();
    sSource.Replace("::", "_");
    sSource.ToLower();
    sSource += ".cpp";

    return sSource;
}

s_str Module::GetCreatedHeaderFile() const
{
    s_str sHeader = GetFullName();
    sHeader.Replace("::", "_");
    sHeader.ToLower();
    sHeader += ".hpp";

    return sHeader;
}

void Module::UsesModule(s_wptr<Module> pModule)
{
    if (lUsedModuleList_.Find(pModule))
    {
        pCompiler_->ThrowWarning("module \'" + pModule->GetFullName() + "\' has already been flagged as used, ignoring");
        return;
    }

    lUsedModuleList_.PushBack(pModule);
}

void Module::UsesNamespace(const s_ctnr<s_str>& lNamespace)
{
    // TODO : check namespace
    lUsedNamespaceList_.PushBack(lNamespace);
}

void Module::UsesType(const s_ctnr<s_str>& lType)
{
    // TODO : check type
    TypeName mType;
    lUsedTypeList_.PushBack(mType);
}

void Module::UsesContent(const s_str& sContent)
{
    lInternalContent_.PushBack(Beautify_(sContent));
}

void Module::Exports(const s_str& sContent)
{
    lExportedContent_.PushBack(Beautify_(sContent));
}

void Module::SetScope(const s_str& sScope)
{
    if (sScope != "public" && sScope != "private" && sScope != "protected")
        pCompiler_->ThrowError("unknown scope '" + sScope + "'");

    if (pCurrentClass_)
        pCurrentClass_->SetScope(sScope);
    else
        pCompiler_->ThrowError("cannot set scope when not inside a class or struct");
}

void Module::CheckType_(const TypeName& mTypeName)
{
    if (!mTypeName.pModule)
        return;

    if (mTypeName.bIsTemplate || !mTypeName.bIsPointer || mTypeName.bIsTypedef || mTypeName.bIsEnum || mTypeName.bIsNested)
    {
        HeaderNeedsModule_(mTypeName.pModule);
    }

    s_ctnr<TypeName>::const_iterator iterTemplate;
    foreach (iterTemplate, mTypeName.lTemplateParams)
        CheckType_(*iterTemplate);
}

s_wptr<Class> Module::BeginClass(const s_str& sName, const s_str& sClassType)
{
    s_wptr<Class> pClass;

    if (!pCurrentClass_)
    {
        s_ctnr<s_refptr<Class>>::const_iterator iterClass;
        foreach (iterClass, mGlobalScope_.lClassList_)
        {
            if ((*iterClass)->GetType().sType == sName)
                pCompiler_->ThrowError("a class with the name \'" + sName + "\' already exists in global scope");
        }

        mGlobalScope_.lClassList_.PushBack(s_refptr<Class>(new Class(pCompiler_, pSelf_, sClassType)));
        pClass = mGlobalScope_.lClassList_.Back();

        TypeName mName;
        mName.sType = sName;
        mName.sFullType = (mName_.sNamespace.IsEmpty() ? sName : mName_.sNamespace + "::" + sName);
        mName.sNakedType = sName;
        mName.sFullNakedType = mName.sFullType;
        mName.sNamespace = mName_.sNamespace;
        mName.lNamespaces = mName_.lNamespaces;
        mName.sNestedType = mName.sType;
        mName.pClass = pClass;

        pClass->SetType(mName);
    }
    else
        pClass = pCurrentClass_->AddClass(sName, sClassType);

    lClassStack_.PushBack(pClass);
    pCurrentClass_ = lClassStack_.Back();

    const TypeName& mType = pClass->GetType();

    pCurrentContext_->lTypeList.PushBack(mType);
    if (sClassType != "union")
        pCurrentContext_->lClassList.PushBack(mType);

    pCurrentContext_->lContextList[mType.sType] = Context(pCurrentContext_);
    pCurrentContext_ = pCurrentContext_->lContextList[mType.sType];

    return pClass;
}

void Module::EndClass()
{
    lClassStack_.PopBack();

    if (lClassStack_.IsEmpty())
        pCurrentClass_ = pMainClass_;
    else
        pCurrentClass_ = lClassStack_.Back();

    pCurrentContext_ = pCurrentContext_.pParent;
}

s_wptr<Class> Module::GetCurrentClass()
{
    return pCurrentClass_;
}

void Module::AddVariable(Variable mVariable)
{
    if (!bPreloaded_)
        CheckType_(mVariable.mType);

    if (mVariable.bGlobal || !pCurrentClass_)
    {
        if (!pCurrentClass_ && mVariable.bGlobal)
            pCompiler_->ThrowWarning("'global' qualifier is useless here", true);

        s_ctnr<Variable>::iterator iterVar;
        foreach (iterVar, mGlobalScope_.lVariableList_)
        {
            if (iterVar->sName == mVariable.sName)
                pCompiler_->ThrowError("a variable with the name \'" + mVariable.sName + "\' already exists in global scope");
        }

        mGlobalScope_.lVariableList_.PushBack(mVariable);
        return;
    }

    pCurrentClass_->AddVariable(mVariable);
}

void Module::AddFunction(Function mFunction)
{
    if (!bPreloaded_)
    {
        CheckType_(mFunction.mReturnType);

        s_ctnr<Variable>::const_iterator iterVar;
        foreach (iterVar, mFunction.lParams)
            CheckType_(iterVar->mType);

        if (!mFunction.lDependencies.IsEmpty())
        {
            s_ctnr<TypeName>::const_iterator iterDeps;
            foreach (iterDeps, mFunction.lDependencies)
            {
                s_bool bFound;
                s_ctnr<s_wptr<Module>>::iterator iterModule;
                foreach (iterModule, lUsedModuleList_)
                {
                    if ((*iterModule)->GetFullName() == iterDeps->sFullType)
                    {
                        bFound = true;
                        HeaderNeedsModule_(*iterModule);
                        break;
                    }
                }

                if (!bFound)
                    pCompiler_->ThrowError("module '"+iterDeps->sFullType+"' has not been previously declared as used");
            }
        }
    }

    if (mFunction.bGlobal || !pCurrentClass_)
    {
        if (mFunction.bGlobal && !pCurrentClass_)
            pCompiler_->ThrowWarning("'global' qualifier is useless here", true);

        mGlobalScope_.lFunctionList_.PushBack(mFunction);
        return;
    }

    pCurrentClass_->AddFunction(mFunction);
}

void Module::AddTypedef(Typedef mTypedef)
{
    if (!bPreloaded_)
    {
        if (mTypedef.bFunction)
        {
            CheckType_(mTypedef.mReturnType);

            if (!mTypedef.mMembership.sType.IsEmpty())
                CheckType_(mTypedef.mMembership);

            s_ctnr<TypeName>::const_iterator iterParams;
            foreach (iterParams, mTypedef.lParams)
                CheckType_(*iterParams);
        }
        else
            CheckType_(mTypedef.mOldType);
    }

    if (mTypedef.bGlobal || !pCurrentClass_)
    {
        if (!pCurrentClass_ && mTypedef.bGlobal)
            pCompiler_->ThrowWarning("'global' qualifier is useless here", true);

        s_ctnr<Typedef>::iterator iterType;
        foreach (iterType, mGlobalScope_.lTypedefList_)
        {
            if (iterType->mType.sType == mTypedef.mType.sType)
                pCompiler_->ThrowError("a typedef with the name \'" + mTypedef.mType.sType + "\' already exists in global scope");
        }

        mGlobalScope_.lTypedefList_.PushBack(mTypedef);

        TypeName& mName = mGlobalScope_.lTypedefList_.Back().mType;
        mName.sFullType = mName_.sNamespace.IsEmpty() ? mName.sType : mName_.sNamespace + "::" + mName.sType;
        mName.sNakedType = mName.sType;
        mName.sFullNakedType = mName.sFullType;
        mName.sNamespace = mName_.sNamespace;
        mName.lNamespaces = mName_.lNamespaces;
        mName.sNestedType = mName.sType;
        mName.bIsTypedef = true;

        pCurrentContext_->lTypeList.PushBack(mName);

        return;
    }

    s_ptr<Typedef> pTypedef = pCurrentClass_->AddTypedef(mTypedef);
    pCurrentContext_->lTypeList.PushBack(pTypedef->mType);

}

void Module::AddEnum(Enum mEnum)
{
    if (mEnum.bGlobal || !pCurrentClass_)
    {
        if (!pCurrentClass_ && mEnum.bGlobal)
            pCompiler_->ThrowWarning("'global' qualifier is useless here", true);

        s_ctnr<Enum>::iterator iterEnum;
        foreach (iterEnum, mGlobalScope_.lEnumList_)
        {
            if (iterEnum->mType.sType == mEnum.mType.sType)
                pCompiler_->ThrowError("an enum with the name \'" + mEnum.mType.sType + "\' already exists in global scope");
        }

        mGlobalScope_.lEnumList_.PushBack(mEnum);

        TypeName& mName = mGlobalScope_.lEnumList_.Back().mType;
        mName.sFullType = mName_.sNamespace.IsEmpty() ? mName.sType : mName_.sNamespace + "::" + mName.sType;
        mName.sNakedType = mName.sType;
        mName.sFullNakedType = mName.sFullType;
        mName.sNamespace = mName_.sNamespace;
        mName.lNamespaces = mName_.lNamespaces;
        mName.sNestedType = mName.sType;
        mName.bIsEnum = true;

        pCurrentContext_->lTypeList.PushBack(mName);

        return;
    }

    s_ptr<Enum> pEnum = pCurrentClass_->AddEnum(mEnum);
    pCurrentContext_->lTypeList.PushBack(pEnum->mType);

}

void Module::AddInheritance(TypeName mClass)
{
    CheckType_(mClass);
    pCurrentClass_->AddInheritance(mClass);
}

s_ctnr<TypeName> Module::DecorateType_(const TypeName& mType, const s_ctnr<s_str>& lContext) const
{
    s_ctnr<TypeName> lList;
    TypeName mTemp;

    if (pMainClass_)
        lList = pMainClass_->DecorateType(mType, lContext);
    else
    {
        s_ctnr<s_refptr<Class>>::const_iterator iterClass;
        foreach (iterClass, mGlobalScope_.lClassList_)
        {
            s_ctnr<TypeName> lTempList = (*iterClass)->DecorateType(mType, lContext);
            if (!lTempList.IsEmpty())
                lList.PushBack(lTempList);
        }
    }

    s_ctnr<Typedef>::const_iterator iterType;
    foreach (iterType, mGlobalScope_.lTypedefList_)
    {
        if (iterType->mType.TryDecorateType(mType, mTemp, pSelf_, lContext))
            lList.PushBack(mTemp);
    }
    s_ctnr<Enum>::const_iterator iterEnum;
    foreach (iterEnum, mGlobalScope_.lEnumList_)
    {
        if (iterEnum->mType.TryDecorateType(mType, mTemp, pSelf_, lContext))
            lList.PushBack(mTemp);
    }

    return lList;
}

s_bool Module::IsReachable_(const TypeName& mType) const
{
    // FIXME : IsReachable seems not to be working
    // FIXME : allow special case (TypeName.bOwner) :
    /*
    class X
    {
    private :

        class Y
        {

        };

        class Z
        {
            Y y;
        };
    };
    */

    /*if (!mType.bIsProtected && !mType.bIsPrivate)
        return true;

    if (!pCurrentClass_)
        return false;

    if (mType.bIsPrivate)
        return (mType.pClass == pCurrentClass_);

    return (mType.pClass == pCurrentClass_ || pCurrentClass_->InheritsFrom(mType.pClass));*/

    // TODO : check reachable in TryDecorateType
    // pretty error messages will have to be dealt with later on
    // by addind a "possible candidate" list
    // and allow "foo is private in this context"
    return true;
}

void Module::DecorateType(TypeName& mType/*, const s_ctnr<s_str>& lContext*/) const
{
    s_str sInitial = mType.sType;
    s_ctnr<TypeName> lList;
    TypeName mTemp;

    s_ctnr<s_str> lLocalContext;
    s_ctnr<s_str> lWords = mType.sType.Cut("::");
    s_ctnr<s_str>::iterator iter;
    foreach (iter, lWords)
        lLocalContext.PushBack(lWords);

    lLocalContext.PopBack();

    if (lLocalContext.IsEmpty())
    {

    }

    s_ctnr<s_str> lContext;
    s_ptr<Context> pContext = pCurrentContext_;
    while (pContext->pParent)
    {
        if (lLocalContext.Front() == pContext->sName)
            break;

        lContext.PushFront(pContext->sName);
        pContext = pContext->pParent;
    }


    /*s_str sInitial = mType.sNakedType;
    s_ctnr<TypeName> lList;
    TypeName mTemp;

    if (pMainClass_)
        lList = pMainClass_->DecorateType(mType, lContext);
    else
    {
        s_ctnr<s_refptr<Class>>::const_iterator iterClass;
        foreach (iterClass, mGlobalScope_.lClassList_)
        {
            s_ctnr<TypeName> lTempList = (*iterClass)->DecorateType(mType, lContext);
            if (!lTempList.IsEmpty())
                lList.PushBack(lTempList);
        }
    }

    s_ctnr<Typedef>::const_iterator iterType;
    foreach (iterType, mGlobalScope_.lTypedefList_)
    {
        if (iterType->mType.TryDecorateType(mType, mTemp, pSelf_, lContext))
            lList.PushBack(mTemp);
    }
    s_ctnr<Enum>::const_iterator iterEnum;
    foreach (iterEnum, mGlobalScope_.lEnumList_)
    {
        if (iterEnum->mType.TryDecorateType(mType, mTemp, pSelf_, lContext))
            lList.PushBack(mTemp);
    }

    s_ctnr<s_wptr<Module>>::const_iterator iterModule;
    foreach (iterModule, lUsedModuleList_)
        lList.PushBack((*iterModule)->DecorateType_(mType, lContext));

    if (lList.GetSize() > 1)
    {
        s_str sError;
        s_ctnr<TypeName>::iterator iter;
        foreach (iter, lList)
        {
            if (!IsReachable_(*iter))
                continue;

            if (iter == lList.Begin())
                sError += iter->sFullNakedType;
            else
                sError += ", " + iter->sFullNakedType;
        }

        if (sError.IsEmpty())
        {
            sError = "unreachable type : '"+sInitial+"', available possibilities :\n      ";
            foreach (iter, lList)
            {
                if (iter == lList.Begin())
                    sError += iter->sFullNakedType + (iter->bIsPrivate ? " (private)" : " (protected)");
                else
                    sError += ", " + iter->sFullNakedType + (iter->bIsPrivate ? " (private)" : " (protected)");
            }

            pCompiler_->ThrowError(sError);
        }

        sError = "ambiguous type '"+sInitial+"', available possibilities :\n      "+sError;

        pCompiler_->ThrowError(sError);
    }

    if (lList.GetSize() == 1)
    {
        mTemp = lList.Back();

        if (IsReachable_(mTemp))
            mType = lList.Back();
        else
            pCompiler_->ThrowError("\'"+sInitial+"\' ("+mTemp.sFullNakedType+") is "+s_str(mTemp.bIsPrivate ? "private" : "protected"));
    }
    else
        pCompiler_->ThrowError("unknown type : '"+sInitial+"'");*/
}

s_bool bVerbose = false;

void Module::Serialize() const
{
    if (GetFullName() != "main")
    {
        if (bVerbose) Log("  header... : "+ pCompiler_->GetOutputDirectory() + GetCreatedHeaderFile());
        File mHeader(pCompiler_->GetOutputDirectory() + GetCreatedHeaderFile(), File::O);

        s_str sPreproc = GetFullName();
        sPreproc.Replace("::", "_");
        sPreproc.ToUpper();

        // Header guards
        if (bVerbose) Log("   guards...");
        mHeader.WriteLine("#ifndef " + sPreproc + "_INCLUDED");
        mHeader.WriteLine("#define " + sPreproc + "_INCLUDED");

        // Includes
        if (bVerbose) Log("   includes...");
        s_bool bFirst = true;
        s_ctnr<s_wptr<Module>>::const_iterator iterModule;
        foreach (iterModule, lUsedModuleList_)
        {
            if (!lExternalModuleList_.Find((*iterModule)->GetFullName()))
                continue;

            if (bFirst)
            {
                mHeader.WriteLine("");
                bFirst = false;
            }

            mHeader.WriteLine("#include " + (*iterModule)->GetHeaderFile());
        }

        // Exports
        if (bVerbose) Log("   exports...");
        if (!lExportedContent_.IsEmpty())
        {
            mHeader.WriteLine("");

            s_ctnr<s_str>::const_iterator iterContent;
            foreach (iterContent, lExportedContent_)
            {
                mHeader.WriteLine(*iterContent);
            }
        }

        // Forward declarations
        if (bVerbose) Log("   forward declarations...");
        bFirst = true;
        foreach (iterModule, lUsedModuleList_)
        {
            if (lExternalModuleList_.Find((*iterModule)->GetFullName()))
                continue;

            const s_str& sDecl = (*iterModule)->GetForwardDeclarations();
            if (!sDecl.IsEmpty())
            {
                if (bFirst)
                {
                    mHeader.WriteLine("");
                    bFirst = false;
                }

                mHeader.WriteLine(sDecl);
            }
        }

        // Namespaces
        if (bVerbose) Log("   namespaces...");
        if (!mName_.lNamespaces.IsEmpty())
        {
            mHeader.WriteLine("");

            s_ctnr<s_str>::const_iterator iter;
            foreach (iter, mName_.lNamespaces)
            {
                mHeader.WriteLine("namespace " + *iter + " {");
            }
        }

        // Serialize main class
        if (pMainClass_)
        {
            if (bVerbose) Log("   main class...");
            SerializeClassHeader_(mHeader, *pMainClass_);
        }
        else
        {
            if (bVerbose) Log("   classes...");
            s_ctnr<s_refptr<Class>>::const_iterator iterClass;
            foreach (iterClass, mGlobalScope_.lClassList_)
            {
                SerializeClassHeader_(mHeader, *(*iterClass));
            }
        }

        // Global scope
        if (bVerbose) Log("   globals...");

        if (!mGlobalScope_.lEnumList_.IsEmpty())
        {
            mHeader.WriteLine("");

            s_ctnr<Enum>::const_iterator iterEnum;
            foreach (iterEnum, mGlobalScope_.lEnumList_)
            {
                const Enum& e = *iterEnum;

                s_str sDoc = SerializeDocumentation_(e.lDocElements);
                if (!sDoc.IsEmpty()) mHeader.WriteLine(sDoc);

                mHeader.WriteLine("enum "+e.mType.sType+"\n{\n"+Beautify_(e.sContent, 1)+"\n};");
            }
        }

        if (!mGlobalScope_.lTypedefList_.IsEmpty())
        {
            mHeader.WriteLine("");

            s_ctnr<Typedef>::const_iterator iterType;
            foreach (iterType, mGlobalScope_.lTypedefList_)
                mHeader.WriteLine(SerializeTypedef_(*iterType));
        }

        if (!mGlobalScope_.lVariableList_.IsEmpty())
        {
            mHeader.WriteLine("");

            s_ctnr<Variable>::const_iterator iterVar;
            foreach (iterVar, mGlobalScope_.lVariableList_)
            {
                const Variable& v = *iterVar;
                s_str sDoc = SerializeDocumentation_(v.lDocElements);
                if (!sDoc.IsEmpty()) mHeader.WriteLine(sDoc);
                mHeader.WriteLine("extern " + v.mType.sFullType + " " + v.sName + ";");
            }
        }

        if (!mGlobalScope_.lFunctionList_.IsEmpty())
        {
            s_ctnr<Function>::const_iterator iterFunc;
            foreach (iterFunc, mGlobalScope_.lFunctionList_)
            {
                const Function& f = *iterFunc;

                mHeader.WriteLine("");
                s_str sDoc = SerializeDocumentation_(f.lDocElements);
                if (!sDoc.IsEmpty()) mHeader.WriteLine(sDoc);
                mHeader.WriteLine(SerializeFunctionHeader_(f));
            }
        }

        // Close namespaces
        if (bVerbose) Log("   close namespaces...");
        if (!mName_.lNamespaces.IsEmpty())
        {
            mHeader.WriteLine("");

            s_str sNamespaces;
            s_ctnr<s_str>::const_iterator iter;
            foreach (iter, mName_.lNamespaces)
            {
                sNamespaces += "}\n";
            }

            mHeader.WriteLine(sNamespaces);
        }

        // Close header guards
        if (bVerbose) Log("   close header guards...");
        mHeader.WriteLine("#endif");

        mHeader.Close();
        if (bVerbose) Log("  done.");
    }

    if (bVerbose) Log("  source ...");
    File mSource(pCompiler_->GetOutputDirectory() + GetCreatedSourceFile(), File::O);

    // Include its header
    if (GetFullName() != "main")
        mSource.WriteLine("#include " + GetHeaderFile());

    // Include other headers
    s_ctnr<s_wptr<Module>>::const_iterator iterModule;
    if (!lUsedModuleList_.IsEmpty())
    {
        s_bool bFirst = (GetFullName() != "main");
        foreach (iterModule, lUsedModuleList_)
        {
            if (!lExternalModuleList_.Find((*iterModule)->GetFullName()))
            {
                if (bFirst)
                {
                    mSource.WriteLine("\n#include " + (*iterModule)->GetHeaderFile());
                    bFirst = false;
                }
                else
                    mSource.WriteLine("#include " + (*iterModule)->GetHeaderFile());
            }
        }
    }

    // "using" directives
    foreach (iterModule, lUsedModuleList_)
    {
        if ((*iterModule)->GetName() == GetName())
        {
            pCompiler_->ThrowWarning("ambiguity between '"+GetFullName()+"' and '"+
                (*iterModule)->GetFullName()+"', cannot use 'using "+
                (*iterModule)->GetFullName()+"' C++ directive", true);
            continue;
        }

        const s_str& sDir = (*iterModule)->GetUsingDirectives();
        if (!sDir.IsEmpty())
            mSource.WriteLine("\n"+sDir);
    }

    // "using namespace" directives
    if (!mName_.lNamespaces.IsEmpty())
    {
        s_str sNamespace;
        s_ctnr<s_str>::const_iterator iter;
        foreach (iter, mName_.lNamespaces)
        {
            if (iter == mName_.lNamespaces.Begin())
                sNamespace += *iter;
            else
                sNamespace += "::" + *iter;
        }
        mSource.WriteLine("\nusing namespace " + sNamespace + ";");
    }

    // Internal content
    if (!lInternalContent_.IsEmpty())
    {
        mSource.WriteLine("");

        s_ctnr<s_str>::const_iterator iterContent;
        foreach (iterContent, lInternalContent_)
        {
            s_str sContent = *iterContent;
            mSource.WriteLine(*iterContent);
        }
    }

    // Global variables
    if (!mGlobalScope_.lVariableList_.IsEmpty())
    {
        mSource.WriteLine("");

        s_ctnr<Variable>::const_iterator iterVar;
        foreach (iterVar, mGlobalScope_.lVariableList_)
        {
            const Variable& v = *iterVar;
            if (v.sValue.IsEmpty())
                mSource.WriteLine(v.mType.sNestedType + " " + v.sName + ";");
            else
                mSource.WriteLine(v.mType.sNestedType + " " + v.sName + " = " + v.sValue + ";");
        }
    }

    // Static variables
    s_ctnr<s_ptr<const Class>> lClassList;
    if (pMainClass_)
        lClassList = pMainClass_->GetClassList();
    else
    {
        s_ctnr<s_refptr<Class>>::const_iterator iterClass;
        foreach (iterClass, mGlobalScope_.lClassList_)
            lClassList.PushBack((*iterClass)->GetClassList());
    }

    if (!lClassList.IsEmpty())
    {
        s_bool bFirst = true;
        s_ctnr<s_ptr<const Class>>::iterator iterClass;
        foreach (iterClass, lClassList)
        {
            s_ptr<const Class> c = *iterClass;
            const s_ctnr<Scope>& lScopeList = c->GetScopeList();
            s_ctnr<Scope>::const_iterator iterScope;
            foreach (iterScope, lScopeList)
            {
                const Scope& mScope = *iterScope;
                if (!mScope.lVariableList_.IsEmpty())
                {
                    s_ctnr<Variable>::const_iterator iterVar;
                    foreach (iterVar, mScope.lVariableList_)
                    {
                        const Variable& v = *iterVar;
                        if (v.bStatic && !v.sValue.IsEmpty())
                        {
                            if (bFirst)
                            {
                                mSource.WriteLine("");
                                bFirst = false;
                            }

                            mSource.WriteLine(
                                v.mType.sNestedType + " " + c->GetType().sNestedType + "::"
                                + v.sName + " = " + v.sValue + ";"
                            );
                        }
                    }
                }
            }
        }

        // Member functions (public, private and protected scopes)
        foreach (iterClass, lClassList)
        {
            SerializeClassSource_(mSource, *(*iterClass));
        }
    }

    // Free functions
    if (!mGlobalScope_.lFunctionList_.IsEmpty())
    {
        s_ctnr<Function>::const_iterator iterFunc;
        foreach (iterFunc, mGlobalScope_.lFunctionList_)
        {
            const Function& f = *iterFunc;
            mSource.WriteLine("\n" + f.mReturnType.sNestedType + " " + f.sName + SerializeFunctionParamsAndBody_(f));
        }
    }
    if (bVerbose) Log("  done.");
}

void Module::SerializeClassHeader_(File& mHeader, const Class& c) const
{
    const s_ctnr<Scope>& lScopeList = c.GetScopeList();
    const s_ctnr<s_str>& lDocElements = c.GetDocElements();
    const TypeName&      mName = c.GetType();

    if (bVerbose) Log("    "+c.GetClassType()+" : "+mName.sFullType+" ...");

    // Inheritance
    s_str sInheritance;
    s_ctnr<Scope>::const_iterator iterScope;

    if (c.GetClassType() != "union")
    {
        foreach (iterScope, lScopeList)
        {
            const Scope& mScope = *iterScope;
            s_ctnr<TypeName>::const_iterator iterInh;
            foreach (iterInh, mScope.lInheritanceList_)
            {
                if (sInheritance.IsEmpty())
                    sInheritance += mScope.sName_ + " " + iterInh->sFullType;
                else
                    sInheritance += ", " + mScope.sName_ + " " + iterInh->sFullType;
            }
        }
    }

    mHeader.WriteLine("");

    // Documentation
    s_str sDoc = SerializeDocumentation_(lDocElements);
    if (!sDoc.IsEmpty())
        mHeader.WriteLine(sDoc);

    // Template params
    if (c.IsTemplate())
        mHeader.WriteLine(c.GetTemplateParams().sSerialized);

    // Class name
    if (sInheritance.IsEmpty())
        mHeader.WriteLine(c.GetClassType() + " " + mName.sType);
    else
        mHeader.WriteLine(c.GetClassType() + " " + mName.sType + " : " + sInheritance);

    mHeader.WriteLine("{");

    // Public, private and protected scopes
    s_bool bFirstScope = true;
    s_bool bWritten = false;
    foreach (iterScope, lScopeList)
    {
        const Scope& mScope = *iterScope;
        if (c.GetClassType() != "union")
        {
            if (bFirstScope)
            {
                if ((c.GetClassType() == "struct" && mScope.sName_ != "public") ||
                    (c.GetClassType() == "class" && mScope.sName_ != "private"))
                {
                    mHeader.WriteLine(mScope.sName_ + " :");
                    bWritten = true;
                }
            }
            else
            {
                mHeader.WriteLine("\n"+mScope.sName_ + " :");
                bWritten = true;
            }
        }

        bFirstScope = false;
        mHeader.AddTab();

        // Enums
        if (!mScope.lEnumList_.IsEmpty())
        {
            if (bWritten)
                mHeader.WriteLine("");
            bWritten = true;

            s_ctnr<Enum>::const_iterator iterEnum;
            foreach (iterEnum, mScope.lEnumList_)
            {
                const Enum& e = *iterEnum;

                s_str sDoc = SerializeDocumentation_(e.lDocElements);
                if (!sDoc.IsEmpty()) mHeader.WriteLine(sDoc);

                mHeader.WriteLine("enum "+e.mType.sType+"\n{\n"+Beautify_(e.sContent, 1)+"\n};");
            }
        }

        // Typedefs
        if (!mScope.lTypedefList_.IsEmpty())
        {
            if (bWritten)
                mHeader.WriteLine("");
            bWritten = true;

            s_ctnr<Typedef>::const_iterator iterType;
            foreach (iterType, mScope.lTypedefList_)
                mHeader.WriteLine(SerializeTypedef_(*iterType));
        }

        // Nested classes
        if (!mScope.lClassList_.IsEmpty())
        {
            bWritten = true;

            s_ctnr<s_refptr<Class>>::const_iterator iterClass;
            foreach (iterClass, mScope.lClassList_)
            {
                SerializeClassHeader_(mHeader, *(*iterClass));
            }
        }

        if (!mScope.lFunctionList_.IsEmpty())
        {
            // Functions (not static)
            s_ctnr<Function>::const_iterator iterFunc;
            foreach (iterFunc, mScope.lFunctionList_)
            {
                const Function& f = *iterFunc;
                if (f.bStatic)
                    continue;

                if (bWritten)
                    mHeader.WriteLine("");
                bWritten = true;

                s_str sDoc = SerializeDocumentation_(f.lDocElements);
                if (!sDoc.IsEmpty()) mHeader.WriteLine(sDoc);
                mHeader.WriteLine(SerializeFunctionHeader_(f, c.IsTemplate()));
            }

            // Functions (static)
            foreach (iterFunc, mScope.lFunctionList_)
            {
                const Function& f = *iterFunc;
                if (!f.bStatic)
                    continue;

                if (bWritten)
                    mHeader.WriteLine("");
                bWritten = true;

                s_str sDoc = SerializeDocumentation_(f.lDocElements);
                if (!sDoc.IsEmpty()) mHeader.WriteLine(sDoc);
                mHeader.WriteLine(SerializeFunctionHeader_(f, c.IsTemplate()));
            }
        }

        if (!mScope.lVariableList_.IsEmpty())
        {
            s_ctnr<s_ptr<const Variable>> lVarList;
            s_ctnr<s_ptr<const Variable>> lStatVarList;

            s_ctnr<Variable>::const_iterator iterVar1;
            foreach (iterVar1, mScope.lVariableList_)
            {
                const Variable& v = *iterVar1;
                if (v.bStatic)
                    lStatVarList.PushBack(&v);
                else
                    lVarList.PushBack(&v);
            }

            if (!lVarList.IsEmpty() && bWritten)
            {
                mHeader.WriteLine("");
                bWritten = true;
            }

            // Variables (not static)
            s_ctnr<s_ptr<const Variable>>::const_iterator iterVar2;
            foreach (iterVar2, lVarList)
            {
                const Variable& v = *(*iterVar2);

                s_str sDoc = SerializeDocumentation_(v.lDocElements);
                if (!sDoc.IsEmpty()) mHeader.WriteLine(sDoc);
                mHeader.WriteLine(v.mType.sFullType + " " + v.sName + ";");
            }

            if (!lStatVarList.IsEmpty() && bWritten)
                mHeader.WriteLine("");

            // Variables (static)
            foreach (iterVar2, lStatVarList)
            {
                const Variable& v = *(*iterVar2);

                s_str sDoc = SerializeDocumentation_(v.lDocElements);
                if (!sDoc.IsEmpty()) mHeader.WriteLine(sDoc);
                mHeader.WriteLine("static " + v.mType.sFullType + " " + v.sName + ";");
            }
        }

        mHeader.RemoveTab();
    }

    mHeader.WriteLine("};");

    if (bVerbose) Log("    done.");
}

void Module::SerializeClassSource_(File& mSource, const Class& c) const
{
    if (c.GetClassType() == "union" || c.IsTemplate())
        return;

    const s_ctnr<Scope>& lScopeList = c.GetScopeList();
    const TypeName&      mName = c.GetType();

    s_ctnr<Scope>::const_iterator iterScope;
    foreach (iterScope, lScopeList)
    {
        const Scope& mScope = *iterScope;
        if (!mScope.lFunctionList_.IsEmpty())
        {
            s_ctnr<Function>::const_iterator iterFunc;
            foreach (iterFunc, mScope.lFunctionList_)
            {
                const Function& f = *iterFunc;
                if (f.bPureVirtual)
                    continue;
                if (!f.mTemplateParams.sSerialized.IsEmpty())
                    continue;

                s_str sFunc = "\n";
                if (f.bConstructor)
                    sFunc += mName.sNestedType + "::" + mName.sType;
                else if (f.bDestructor)
                    sFunc += mName.sNestedType + "::~" + mName.sType;
                else
                    sFunc += f.mReturnType.sNestedType + " " + mName.sNestedType + "::" + f.sName;

                sFunc += SerializeFunctionParamsAndBody_(f);
                mSource.WriteLine(sFunc);
            }
        }
    }
}

s_str Module::Beautify_(const s_str& sStr, const s_uint& uiIndent) const
{
    s_str sResult = sStr;
    while (sResult.Front() == '\n') sResult.PopFront();
    while (sResult.Back()  == '\n') sResult.PopBack();
    sResult.Replace("\t", "    ");

    s_uint uiMinIndent = s_uint::INF;
    s_str::iterator iter;
    s_str sLine;
    s_ctnr<s_str> lLines;
    foreach (iter, sResult)
    {
        if (*iter != '\n')
            sLine.PushBack(*iter);
        else
        {
            s_uint uiIndent;
            s_str::iterator iterLine = sLine.Begin();
            while (iterLine != sLine.End() && (*iterLine) == ' ')
            {
                ++uiIndent;
                ++iterLine;
            }

            if (iterLine != sLine.End())
                uiMinIndent = s_uint::Min(uiMinIndent, uiIndent);

            lLines.PushBack(sLine);
            sLine = "";
        }
    }

    if (!sLine.IsEmpty())
    {
        s_uint uiIndent;
        s_str::iterator iterLine = sLine.Begin();
        while (iterLine != sLine.End() && (*iterLine) == ' ')
        {
            ++uiIndent;
            ++iterLine;
        }

        if (iterLine != sLine.End())
            uiMinIndent = s_uint::Min(uiMinIndent, uiIndent);

        lLines.PushBack(sLine);
        sLine = "";
    }

    s_str sIndent;
    if (uiIndent != 0)
    {
        if (uiIndent*4 > uiMinIndent)
        {
            sIndent = s_str(' ', uiIndent*4 - uiMinIndent);
            uiMinIndent = 0;
        }
        else
            uiMinIndent -= uiIndent*4;
    }
    else if (uiMinIndent == 0)
        return sResult;


    s_ctnr<s_str>::iterator iterLine = lLines.End();
    while (iterLine != lLines.Begin())
    {
        --iterLine;

        if (iterLine->IsEmpty(true))
            iterLine = lLines.Erase(iterLine);
    }

    sResult = "";
    foreach (iterLine, lLines)
    {
        if (!iterLine->IsEmpty(true))
        {
            if (uiMinIndent != 0)
                iterLine->EraseFromStart(uiMinIndent);
            else
                iterLine->PushFront(sIndent);
        }

        if (sResult.IsEmpty())
            sResult = *iterLine;
        else
            sResult += "\n" + *iterLine;
    }

    return sResult;
}

s_str Module::SerializeFunctionHeader_(const Function& f, const s_bool& bTemplate) const
{
    s_str sFunc;

    if (!f.mTemplateParams.sSerialized.IsEmpty())
        sFunc += f.mTemplateParams.sSerialized + "\n";

    if (f.bVirtual)
        sFunc += "virtual ";
    if (f.bStatic)
        sFunc += "static ";

    if (f.bConstructor)
        sFunc += f.sName + "(";
    else if (f.bDestructor)
        sFunc += "~" + f.sName + "(";
    else
        sFunc += f.mReturnType.sFullType + " " + f.sName + "(";

    s_ctnr<Variable>::const_iterator iterParam;
    foreach (iterParam, f.lParams)
    {
        const Variable& v = *iterParam;

        if (iterParam != f.lParams.Begin())
            sFunc += ", ";

        sFunc += v.mType.sFullType + " " + v.sName;

        if (!v.sValue.IsEmpty())
            sFunc += " = " + v.sValue;
    }

    sFunc += ")";

    if (f.bConst)
        sFunc += " const";

    if (f.bPureVirtual)
        sFunc += " = 0";

    if (!f.mTemplateParams.sSerialized.IsEmpty() || bTemplate)
    {
        s_str sContent = Beautify_(f.sContent, 1);

        if (sContent.IsEmpty(true)) sFunc += "\n{\n}";
        else sFunc += "\n{\n" + sContent + "\n}";
    }
    else
        sFunc += ";";

    return sFunc;
}

s_str Module::SerializeFunctionParamsAndBody_(const Function& f) const
{
    s_str sFunc = "(";

    s_ctnr<Variable>::const_iterator iterParam;
    foreach (iterParam, f.lParams)
    {
        const Variable& v = *iterParam;

        if (iterParam != f.lParams.Begin())
            sFunc += ", ";

        sFunc += v.mType.sNestedType + " " + v.sName;
    }

    sFunc += ")";

    if (f.bConstructor && f.lInit.GetSize() != 0)
    {
        sFunc += " : ";
        s_ctnr<Variable>::const_iterator iterVar;
        foreach (iterVar, f.lInit)
        {
            if (iterVar != f.lInit.End()-1)
                sFunc += iterVar->sName + "(" + iterVar->sValue + "), ";
            else
                sFunc += iterVar->sName + "(" + iterVar->sValue + ")";
        }
    }

    if (f.bConst)
        sFunc += " const";

    s_str sContent = Beautify_(f.sContent, 1);

    if (sContent.IsEmpty(true)) sFunc += "\n{\n}";
    else sFunc += "\n{\n" + sContent + "\n}";

    return sFunc;
}

s_str Module::SerializeDocumentation_(const s_ctnr<s_str>& lDocElements) const
{
    s_str sDoc;

    s_ctnr<s_str> lLines;

    s_ctnr<s_str>::const_iterator iter1;
    foreach (iter1, lDocElements)
    {
        if (iter1->IsEmpty(true))
        {
            lLines.PushBack("");
            continue;
        }

        s_ctnr<s_str> lTemp = iter1->Cut("\n");
        s_ctnr<s_str>::const_iterator iter2;
        foreach (iter2, lTemp)
        {
            lLines.PushBack(*iter2);
        }
    }

    foreach (iter1, lLines)
    {
        if (iter1->IsEmpty(true))
            sDoc += "\n";

        s_str sTemp = *iter1;
        sTemp.Trim(' ');

        if (sDoc.IsEmpty())
            sDoc = sTemp;
        else
            sDoc += "\n" + sTemp;
    }

    return sDoc;
}

s_str Module::SerializeTypedef_(const Typedef& t) const
{
    s_str sResult;

    s_str sDoc = SerializeDocumentation_(t.lDocElements);
    if (!sDoc.IsEmpty()) sResult = sDoc+"\n";

    sResult += "typedef ";

    if (t.bFunction)
    {
        sResult += t.mReturnType.sFullType;
        if (t.mMembership.sType.IsEmpty())
            sResult += " (*";
        else
            sResult += " ("+t.mMembership.sFullType+"::*";
        sResult += t.mType.sType + ")(";

        s_ctnr<TypeName>::const_iterator iterParams;
        foreach (iterParams, t.lParams)
        {
            if (iterParams == t.lParams.Begin())
                sResult += iterParams->sFullType;
            else
                sResult += ", "+iterParams->sFullType;
        }

        if (t.bConst)
            sResult += ") const;";
        else
            sResult += ");";
    }
    else
        sResult += t.mOldType.sFullType+" "+t.mType.sType+";";

    return sResult;
}

s_str Module::GetForwardDeclarations() const
{
    s_str sForward;

    s_str sIndent;
    if (!mName_.lNamespaces.IsEmpty())
    {
        s_ctnr<s_str>::const_iterator iter;
        foreach (iter, mName_.lNamespaces)
            sForward += "namespace " + *iter + " {\n";

        sIndent = "    ";
    }

    if (pMainClass_)
        sForward += sIndent + pMainClass_->GetClassType() + " " + mName_.sType + ";";
    else
    {
        if (mGlobalScope_.lClassList_.IsEmpty())
            return "";

        s_ctnr<s_refptr<Class>>::const_iterator iterClass;
        foreach (iterClass, mGlobalScope_.lClassList_)
        {
            if (iterClass != mGlobalScope_.lClassList_.Begin())
                sForward += "\n";

            sForward += sIndent + (*iterClass)->GetClassType() + " " + (*iterClass)->GetType().sType + ";";
        }
    }

    if (!mName_.lNamespaces.IsEmpty())
    {
        s_ctnr<s_str>::const_iterator iter;
        foreach (iter, mName_.lNamespaces)
            sForward += "\n}";
    }

    return sForward;
}

s_str Module::GetHeaderFile() const
{
    if (sHeaderFile_.IsEmpty())
        return "\"" + GetCreatedHeaderFile() + "\"";
    else
        return sHeaderFile_;
}

s_str Module::GetUsingDirectives() const
{
    s_str sUsing;

    if (pMainClass_)
        sUsing = "using " + pMainClass_->GetType().sFullType + ";";
    else
    {
        s_ctnr<s_refptr<Class>>::const_iterator iterClass;
        foreach (iterClass, mGlobalScope_.lClassList_)
        {
            if (!sUsing.IsEmpty())
                sUsing += "\n";

            sUsing += "using " + (*iterClass)->GetType().sFullType + ";";
        }
        s_ctnr<Typedef>::const_iterator iterType;
        foreach (iterType, mGlobalScope_.lTypedefList_)
        {
            if (!sUsing.IsEmpty())
                sUsing += "\n";

            sUsing += "using " + iterType->mType.sFullType + ";";
        }

        s_str sNamespace;
        if (!mName_.sNamespace.IsEmpty())
            sNamespace = mName_.sNamespace + "::";

        s_ctnr<Variable>::const_iterator iterVar;
        foreach (iterVar, mGlobalScope_.lVariableList_)
        {
            if (!sUsing.IsEmpty())
                sUsing += "\n";

            sUsing += "using " + sNamespace + iterVar->sName + ";";
        }
    }

    return sUsing;
}
